package tb3d;

/**
 * <p>Title: TreeBuilder3D</p>
 * <p>Description: Interactive 3D viewer</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: GSC BC Cancer Agency</p>
 *
 * @author Peter Ruzanov
 * @version 1.0
 */

/*GPL
This file is part of TreeBuilder3D

TreeBuilder3D is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

TreeBuilder3D is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeBuilder3D; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


import java.awt.BorderLayout;
//import java.awt.FlowLayout;
import java.awt.event.*;
import java.awt.Font;
import java.awt.Color;
import java.io.FileNotFoundException;
import java.io.RandomAccessFile;

//import java.applet.Applet;
import java.awt.GraphicsConfiguration;
//import java.awt.image.ColorModel;
//import com.sun.j3d.utils.applet.MainFrame;
//import com.sun.j3d.utils.geometry.ColorCube;
import com.sun.j3d.utils.geometry.Sphere;
import com.sun.j3d.utils.geometry.Text2D;


import com.sun.j3d.utils.universe.*;
//import com.sun.j3d.utils.behaviors.mouse.*;
import com.sun.j3d.utils.behaviors.vp.OrbitBehavior;
import javax.swing.*;
import javax.swing.event.*;

import javax.media.j3d.*;
import javax.vecmath.*;
import java.util.Enumeration;
import java.util.Hashtable;
//import java.util.Vector;

public class tb3d extends JPanel implements Runnable{

   // public static void main(String[] args){
   // new MainFrame(new tb3d(),350,280);
   //}

  private SNode [] nodes;
  private Edge edges[];
  //Additional components for NJ:
  private SubNode [] knots;
  private Edge kedges [];
  private Thread relaxer;

  private SimpleUniverse u = null;
  private boolean dataLoaded = false;
  private String[] NAMES = {"EXPRESSION","DATA","ANALYSIS"};
  private String[] DESCS = {"ALTERNATIVE","DESCRIPTION","MODE"};
  private int[][] MATRIX = {{0,123,354},
			    {0,0,200}};
  private int description_length = 25;
  //private boolean[][] v_index; //visibility index

  private JFileChooser fileChooser = new JFileChooser();
  private JSlider VSlider; //visibility slider
  private JSlider DSlider; //density slider
  private JCheckBox c_box; //NJ box
  private JCheckBox d_box; //Description box
  private float coeff = 1.0f;
  private float refcf = 1.0f;


  private Color3f[][] stress = {
			{new Color3f(Color.BLACK),new Color3f(Color.BLACK)},
			{new Color3f(Color.BLUE), new Color3f(Color.BLUE)},
			{new Color3f(new Color(60,60,150)), new Color3f(new Color(60,60,150))},
			{new Color3f(new Color(150,60,60)), new Color3f(new Color(150,60,60))},
			{new Color3f(Color.RED),  new Color3f(Color.RED)}};
/**
 *  CONSTRUCTOR
 */

  public tb3d() {
    this.setLayout(new BorderLayout());
    Canvas3D c = new Canvas3D(SimpleUniverse.getPreferredConfiguration());
    add("Center", c);

    //Sliders:
    VSlider = new JSlider();
    add("North",VSlider);
    VSlider.setMinimum(0);
    VSlider.setMaximum(100);
    VSlider.addChangeListener(
	    new javax.swing.event.ChangeListener() {
		public void stateChanged(ChangeEvent e) {
		    VSlider_stateChanged(e);
		}
	    });

    DSlider = new JSlider();
    DSlider.setOrientation(JSlider.VERTICAL);
    add("West",DSlider);
    DSlider.setMinimum(1);
    DSlider.setMaximum(20);
    DSlider.setValue(5);
    DSlider.addChangeListener(
	    new javax.swing.event.ChangeListener() {
		public void stateChanged(ChangeEvent e) {
		    DSlider_stateChanged(e);
		}
	    });

    //Buttons:
    JToolBar tools = new JToolBar();
    tools.setOrientation(JToolBar.VERTICAL);
    add("East",tools);
    JButton oB = new JButton("Open Matrix");
    tools.add(oB);
    oB.addActionListener(new java.awt.event.ActionListener() {
    public void actionPerformed(ActionEvent e) {
      Open_actionPerformed(e);
      }
      });

    JButton test = new JButton("Shake");
    tools.add(test);
    test.addActionListener(new java.awt.event.ActionListener() {
    public void actionPerformed(ActionEvent e) {
      Test_actionPerformed(e);
      }
      });

    c_box = new JCheckBox("BRANCH",null,false);
    d_box = new JCheckBox("DESCR",null,false);
    tools.add(c_box);
    tools.add(d_box);
    c_box.setFocusPainted(false);
    c_box.setOpaque(false);
    c_box.addActionListener(new java.awt.event.ActionListener() {
    public void actionPerformed(ActionEvent e) {
      Cbox_actionPerformed(e);
      }
      });

    d_box.setFocusPainted(false);
    d_box.setOpaque(false);
    d_box.addActionListener(new java.awt.event.ActionListener() {
    public void actionPerformed(ActionEvent e) {
      Dbox_actionPerformed(e);
      }
      });

    //this.maxlength = 0;
    u = new SimpleUniverse(c);

    ViewingPlatform vp = u.getViewingPlatform();
    OrbitBehavior orbit = new OrbitBehavior(c,OrbitBehavior.REVERSE_ALL | OrbitBehavior.MOUSE_LISTENER);
    BoundingSphere bounds = new BoundingSphere(new Point3d(0.0,0.0,0.0),100.0);
    orbit.setSchedulingBounds(bounds);
    vp.setViewPlatformBehavior(orbit);
    vp.setNominalViewingTransform();

    BranchGroup scene = createSceneGraph();



    u.addBranchGraph(scene);
    this.start();
    try {
      jbInit();
    }
    catch(Exception e) {
      e.printStackTrace();
    }
    }


    public BranchGroup createSceneGraph() {

      Appearance ap = new Appearance();
      Appearance aps = new Appearance();
      Material mat1 = new Material();
      mat1.setCapability(Material.ALLOW_COMPONENT_WRITE);
      mat1.setDiffuseColor( new Color3f( 0.5f, 0.2f, 0.8f ) );
      ap.setMaterial(mat1);
      aps.setMaterial(mat1);


      BranchGroup objRoot = new BranchGroup();
      TransformGroup tGroup = new TransformGroup();
      tGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
      tGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

      objRoot.addChild(tGroup);

      BoundingSphere bounds = new BoundingSphere(new Point3d(0.0,0.0,0.0),10);


       if(this.NAMES!=null && this.NAMES.length>0){
       double maxlength = 0;
       this.edges = new Edge[((this.NAMES.length-1) * (this.NAMES.length-1) + this.NAMES.length-1)/2];
       this.nodes = new SNode[this.NAMES.length];
       //Subnodes and subedges:
       if(this.NAMES.length>=3)
	this.knots = new SubNode[this.NAMES.length - 2];
       if(this.NAMES.length>=2)
	this.kedges = new Edge[1+2*(this.NAMES.length-2)];
       int ordinal = 0;


       for (int i = 0; i < this.NAMES.length; i++){             //NODES
	addNode(this.NAMES[i],i,ap);
	this.nodes[i].bb.setSchedulingBounds(bounds);
	tGroup.addChild(this.nodes[i].tg);
	tGroup.addChild(this.nodes[i].lg);
	tGroup.addChild(this.nodes[i].bb);
	//AFTER ALL DONE WITH NEW CODE WE NEED TO ADD KEDGES AND NODES HERE!


	for (int j=i+1;j<this.NAMES.length;j++){
	if(MATRIX[i][j]<=0)
	   continue;
	this.edges[ordinal] = new Edge(i,j);
	if(MATRIX[i][j] > maxlength){maxlength = MATRIX[i][j];}
	this.edges[ordinal].len  = MATRIX[i][j];
	tGroup.addChild(this.edges[ordinal].line);
	ordinal++;
	}

	this.coeff = (float)(1.0/maxlength);
	this.refcf = coeff;
	this.VSlider.setMaximum((int)maxlength);
	this.VSlider.setValue((int)maxlength);
       }
      //NEW CODE:
      this.buildTree();
      for (int i = 0; i < this.knots.length; i++){             //SUBNODES
	addSNode(i, aps);
	this.knots[i].getAppearance().getRenderingAttributes().setVisible(false);
	tGroup.addChild(this.knots[i].tg);
      }
      for (int j = 0;j < this.kedges.length; j++) {
	  this.kedges[j].line.getAppearance().getRenderingAttributes().setVisible(false);
	  tGroup.addChild(this.kedges[j].line);
	}
      }

      Background bg = new Background(new Color3f(Color.GRAY));  //BACKGOUND
      bg.setApplicationBounds(bounds);



      DirectionalLight light1 = new DirectionalLight();         //LIGHT
      light1.setEnable( true );
      light1.setColor( new Color3f(0.7f, 0.7f, 0.7f) );
      light1.setDirection( new Vector3f( 1.0f, 0.5f, -1.0f ) );
      light1.setInfluencingBounds( bounds );

      DirectionalLight light2 = new DirectionalLight();         //LIGHT
      light2.setEnable( true );
      light2.setColor( new Color3f(0.7f, 0.5f, 0.5f) );
      light2.setDirection( new Vector3f( -1.0f, -0.5f, 1.0f ) );
      light2.setInfluencingBounds( bounds );


      objRoot.addChild(light1);
      objRoot.addChild(light2);
      objRoot.addChild(bg);

      for(int e = 0;e < edges.length;e++){                            //EDGES
	  Point3f first = new Point3f(nodes[edges[e].from].x,nodes[edges[e].from].y,nodes[edges[e].from].z);
	  Point3f secnd = new Point3f(nodes[edges[e].to].x,nodes[edges[e].to].y,nodes[edges[e].to].z);
	  Point3f[] tp = {first,secnd};

	  this.edges[e].la.setCoordinates(0,tp);
      }
      for(int k = 0;k < kedges.length;k++){                          //KEDGES
	float kx,ky,kz;
	int points[] = {kedges[k].from,kedges[k].to};
	Point3f[] tp = new Point3f[2];

	for(int p = 0;p < points.length;p++){
	  if(points[p] >= NAMES.length){
	  kx = knots[points[p] - NAMES.length].x;
	  ky = knots[points[p] - NAMES.length].y;
	  kz = knots[points[p] - NAMES.length].z;
	}else{
	  kx = nodes[points[p]].x;
	  ky = nodes[points[p]].y;
	  kz = nodes[points[p]].z;
	}
	tp[p] = new Point3f(kx,ky,kz);
	}
	this.kedges[k].la.setCoordinates(0,tp);
      }

      objRoot.setCapability(BranchGroup.ALLOW_DETACH);
      objRoot.compile();
      return objRoot;
    }


/**
* Destructor
*/
   public void destroy() {
     u.removeAllLocales();
   }


/**
* Start method: starting the relaxing thread
*
*/
    public void start() {
	       relaxer = new Thread(this);
	       relaxer.start();
    }

/**
*  Main processing method for the TreePanel object
*/
    public void run() {
      Thread me = Thread.currentThread();
      while (relaxer == me) {
	     relax();
	     try {
		 Thread.sleep(100);
	      } catch (InterruptedException e) {
		  break;
	      }
	  }
    }


/**
* Updating graph
*/
    public void updateGraph(String[]NAMES,int[][] grids){
      //Little validation first:
      if(NAMES == null || grids == null)
	return;
      if(this.c_box.isSelected())
	this.c_box.setSelected(false);
      if(this.d_box.isSelected())
	this.d_box.setSelected(false);

      for(int n=0; n<NAMES.length; n++){
      if(NAMES[n] == null)
	return;
      }
	for(int g=0;g<NAMES.length; g++){
	 for(int h=g+1;h<NAMES.length;h++){
	 if(grids[g][h] == 0)
	   return;
	 }
	}
      //Validated data

      Enumeration en = u.getLocale().getAllBranchGraphs();
      BranchGroup obg = null;
      int num = 0;//flags Viewer Platform node, should be changes to more reliable thing
      while(en.hasMoreElements() ){
      obg = (BranchGroup)en.nextElement();

      if(num>0)
	u.getLocale().removeBranchGraph(obg);
      num++;
      }
      u.addBranchGraph(loadDistances(NAMES,grids));
    }

/**
* Function for initialization of nodes on 3D canvas
*/
   private void addNode(String lbl,int pos, Appearance ap){
      SNode n = new SNode(.06f);

	n.id = lbl;
	 if(lbl.length() > description_length)
	    n.id = lbl.substring(0,description_length - 1);
	  else{
	    while(n.id.length() < description_length)
	     n.id += " ";
	  }

	n.tt = new Transform3D();
	n.tl = new Transform3D();
	n.bbg= new TransformGroup();
	n.x = (float)Math.random()*2-1;
	n.y = (float)Math.random()*2-1;
	n.z = (float)Math.random()*2-1;

	n.tt.set(new Vector3f(n.x,n.y,n.z));
	n.tl.set(new Vector3f(n.x,n.y+0.05f,n.z));
	n.tg = new TransformGroup(n.tt);
	n.lg = new TransformGroup(n.tl);
	n.tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
	n.lg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
	n.bbg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

	n.lbl = new Text2D(n.id,new Color3f(Color.ORANGE),"Seriff",18,Font.BOLD);
	n.lbl.setCapability(Text2D.ALLOW_APPEARANCE_READ);
	n.lbl.setCapability(Text2D.ALLOW_APPEARANCE_WRITE);
	n.lbl.getAppearance().setCapability(Appearance.ALLOW_TEXTURE_READ);
	n.lbl.getAppearance().setCapability(Appearance.ALLOW_TEXTURE_WRITE);
	n.lbl.getAppearance().getTexture().setCapability(Texture.ALLOW_IMAGE_READ);
	n.lbl.getAppearance().getTexture().setCapability(Texture.ALLOW_IMAGE_WRITE);
	n.lbl.getAppearance().getTexture().setCapability(Texture.ALLOW_SIZE_READ);
	n.lbl.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
	n.lbl.setCapability(Shape3D.ALLOW_GEOMETRY_WRITE);
	n.setAppearance(ap);

	this.nodes[pos] = n;
	n.tg.addChild(n);
	n.lg.addChild(n.bbg);
	n.bbg.addChild(n.lbl);
	n.bb = new Billboard();
	n.bb.setAlignmentMode(Billboard.ROTATE_ABOUT_POINT);
	n.bb.setRotationPoint(0.0f,0.0f,0.0f);
	n.bb.setTarget(n.bbg);
    }

    /**
     * Function for initialization of knots on 3D canvas
     */
    private void addSNode(int pos, Appearance ap) {

      knots[pos].tt = new Transform3D();
      float [] x = new float[2];
      float [] y = new float[2];
      float [] z = new float[2];
      for(int l = 0; l<2; l++){
      if(knots[pos].Linked[l]>=NAMES.length){
       x[l] = knots[knots[pos].Linked[l]-NAMES.length].x;
       y[l] = knots[knots[pos].Linked[l]-NAMES.length].y;
       z[l] = knots[knots[pos].Linked[l]-NAMES.length].z;
      }else{
       x[l] = nodes[knots[pos].Linked[l]].x;
       y[l] = nodes[knots[pos].Linked[l]].y;
       z[l] = nodes[knots[pos].Linked[l]].z;
      }
      }
      knots[pos].x = (float) (x[0]+x[1])/2;
      knots[pos].y = (float) (y[0]+y[1])/2;
      knots[pos].z = (float) (z[0]+z[1])/2;

      knots[pos].tt.set(new Vector3f(knots[pos].x, knots[pos].y, knots[pos].z));
      knots[pos].tg = new TransformGroup(knots[pos].tt);
      knots[pos].tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      RenderingAttributes ra = new RenderingAttributes();

      ra.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);
      ap.setRenderingAttributes(ra);
      ap.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_READ);
      knots[pos].setAppearance(ap);
      knots[pos].getShape().setCapability(Shape3D.ALLOW_APPEARANCE_READ);
      knots[pos].tg.addChild(knots[pos]);
    }


/**
* Overriden Sphere class, SNode has transform and transformGroup classes attached
* for easier handling
* @param lbl       Holds the label string
* @param tg        TransformGroup object, animation tasks
* @param tt        Transform3D object, manipulation of node position/scale etc.
* @param grabbed   Flags if the node is being dragged
*/
    private class SNode extends Sphere{
      public SNode(float rad){super(rad);}

    Text2D lbl;
    String id;
    TransformGroup tg;//for node itself
    TransformGroup lg;//for label
    TransformGroup bbg;//for billboard
    Transform3D    tt;//for node itself
    Transform3D    tl;//for label

    Billboard bb;
    boolean grabbed;
    float x,y,z;
    float dx,dy,dz;
    }

    /**
    * Overriden Sphere class, SubNode has transform and transformGroup classes attached
    * for easier handling
    * @param lbl       Holds the label string
    * @param tg        TransformGroup object, animation tasks
    * @param tt        Transform3D object, manipulation of node position/scale etc.
    * @param grabbed   Flags if the node is being dragged
    */
    private class SubNode extends Sphere{
     public SubNode(float rad,int x,int y){
      super(rad);
      //MAJOR CHANGES GO HERE
      this.Linked = new int[2];
      this.Linked[0] = x;
      this.Linked[1] = y;
      this.offset = 0;
      }
     int Linked [];
     int offset;
     TransformGroup tg;
     Transform3D    tt;

     float x,y,z;
     float dx,dy,dz;
     }


/**
* Edge class stores the ordinals for connected nodes
* @param from     Ordinal of the first node in connected pair of nodes
* @param to       Ordinal of the second node in connected pair of nodes
* @param len      The length of connection (received from external distance matrix)
* @param visible  Flags if the edge is visible
*
*/
    private class Edge {
      public Edge(int f, int t){this.from =f;
				this.to   =t;
				this.la = new LineArray(2,LineArray.COORDINATES | LineArray.COLOR_3);

				Point3f[] pts = {new Point3f(-.1f,-.3f,.1f),new Point3f(-.1f,.5f,-.3f)};
				this.la.setCoordinates(0,pts);

				Appearance apt = new Appearance();
				RenderingAttributes ra = new RenderingAttributes();
				ra.setCapability(RenderingAttributes.ALLOW_VISIBLE_WRITE);
				apt.setRenderingAttributes(ra);
				apt.setCapability(Appearance.ALLOW_RENDERING_ATTRIBUTES_READ);


				this.la.setColors(0,stress[0]);
				this.la.setCapability(LineArray.ALLOW_COORDINATE_WRITE);
				this.la.setCapability(LineArray.ALLOW_COLOR_WRITE);
				this.line = new Shape3D(la);
				this.line.setAppearance(apt);
				this.line.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
				}
    LineArray la;
    Shape3D line;

    int from, to;
    double len;
    }

/**
* Function for loading the distance data into local array
* Calls for cleaning if there are any nodes already
* @param  names[]      names of the libraries (data sets)
* @param  matrix[][]   distances from the table
*/
protected BranchGroup loadDistances(String[] libs,int[][] matrix){
  if(dataLoaded)
    return null;
  if(NAMES != null && NAMES.length>0){this.NAMES = null;
				      this.nodes = null;
				      }

this.NAMES = libs;
this.MATRIX = matrix;

BranchGroup scene = createSceneGraph();
scene.compile();
dataLoaded = true;

return scene;
}

/**
*  Method for relaxing the graph. Determines how fast the nodes moves,
*  changes color parameters depending on the stress.
*/
    synchronized void relax() {
      if(!dataLoaded)
	return;
      Point3f fr,to;

      for (int i = 0; i < edges.length; i++) {
	    Edge e = edges[i];
	    fr = new Point3f(nodes[e.from].x,nodes[e.from].y,nodes[e.from].z);
	    to = new Point3f(nodes[e.to].x,nodes[e.to].y,nodes[e.to].z);

	    float vx = nodes[e.to].x - nodes[e.from].x;
	    float vy = nodes[e.to].y - nodes[e.from].y;
	    float vz = nodes[e.to].z - nodes[e.from].z;

	    double len = fr.distance(to);
	    len = (len == 0) ? .0001 : len;
	    double f = (e.len*this.coeff - len) / (len * 3);
	    double dx = f * vx;
	    double dy = f * vy;
	    double dz = f * vz;

	    nodes[e.to].dx   +=  dx*this.coeff;
	    nodes[e.to].dy   +=  dy*this.coeff;
	    nodes[e.to].dz   +=  dz*this.coeff;
	    nodes[e.from].dx += -dx*this.coeff;
	    nodes[e.from].dy += -dy*this.coeff;
	    nodes[e.from].dz += -dz*this.coeff;
	}
	//For knots:
	if(this.c_box.isSelected()){
	  for (int j = 0; j < kedges.length; j++) {
	  Edge k = kedges[j];

	  if(k.from>=NAMES.length)
	   fr = new Point3f(knots[k.from-NAMES.length].x,knots[k.from-NAMES.length].y,knots[k.from-NAMES.length].z);
	  else
	   fr = new Point3f(nodes[k.from].x,nodes[k.from].y,nodes[k.from].z);

	  if(k.to>=NAMES.length)
	   to  = new Point3f(knots[k.to-NAMES.length].x,knots[k.to-NAMES.length].y,knots[k.to-NAMES.length].z);
	  else
	   to  = new Point3f(nodes[k.to].x,nodes[k.to].y,nodes[k.to].z);

	    float kx,ky,kz;

	    if(k.from>=NAMES.length){
	      kx = knots[k.from-NAMES.length].x;
	      ky = knots[k.from-NAMES.length].y;
	      kz = knots[k.from-NAMES.length].z;
	    }else{
	      kx = nodes[k.from].x;
	      ky = nodes[k.from].y;
	      kz = nodes[k.from].z;
	    }

	    if(k.to>=NAMES.length){
	      kx -= knots[k.to-NAMES.length].x;
	      ky -= knots[k.to-NAMES.length].y;
	      kz -= knots[k.to-NAMES.length].z;
	    }else{
	      kx -= nodes[k.to].x;
	      ky -= nodes[k.to].y;
	      kz -= nodes[k.to].z;


	    }

	    double len = fr.distance(to);
	    len = (len == 0) ? .0001 : len;
	    double f = (k.len*this.coeff - len) / (len * 3);
	    double dx = f * kx;
	    double dy = f * ky;
	    double dz = f * kz;

	    if(k.from>=NAMES.length){
	      knots[k.from-NAMES.length].dx += -kx*this.coeff;
	      knots[k.from-NAMES.length].dy += -ky*this.coeff;
	      knots[k.from-NAMES.length].dz += -kz*this.coeff;
	    }
	    if(k.to>=NAMES.length){
	      knots[k.to-NAMES.length].dx   +=  kx*this.coeff;
	      knots[k.to-NAMES.length].dy   +=  ky*this.coeff;
	      knots[k.to-NAMES.length].dz   +=  kz*this.coeff;
	    }
	  }
       //Block dedicated to kedges (subnodes are moved according to kedges length)
       for (int l = 0; l < knots.length; l++) {
	   SubNode n = knots[l];
	       n.x += Math.max(-0.1, Math.min(0.1, n.dx));
	       n.y += Math.max(-0.1, Math.min(0.1, n.dy));
	       n.z += Math.max(-0.1, Math.min(0.1, n.dz));
	   if (n.x < -1.0) {
	       n.x = -1.0f;
	   } else if (n.x*this.coeff > 1.0) {
	       n.x = 1.0f;
	   }
	   if (n.y < -1.0) {
	       n.y = -1.0f;
	   } else if (n.y*this.coeff > 1.0) {
	       n.y = 1.0f;
	   }
	   if (n.z < -1.0) {
	       n.z = -1.0f;
	   } else if (n.z*this.coeff > 1.0) {
	       n.z = 1.0f;
	   }
	   n.dx /= 1.1;
	   n.dy /= 1.1;
	   n.dz /= 1.1;
       }
       Transform3D temp = new Transform3D();
       for(int i = 0;i < this.knots.length; i++){

	 temp.set(new Vector3f(this.knots[i].x,this.knots[i].y,this.knots[i].z));
	 this.knots[i].tg.setTransform(temp);
       }
       }

	for (int i = 0; i < nodes.length; i++) {
	    SNode n = nodes[i];
		n.x += Math.max(-0.1, Math.min(0.1, n.dx));
		n.y += Math.max(-0.1, Math.min(0.1, n.dy));
		n.z += Math.max(-0.1, Math.min(0.1, n.dz));
	    if (n.x < -1.0) {
		n.x = -1.0f;
	    } else if (n.x*this.coeff > 1.0) {
		n.x = 1.0f;
	    }
	    if (n.y < -1.0) {
		n.y = -1.0f;
	    } else if (n.y*this.coeff > 1.0) {
		n.y = 1.0f;
	    }
	    if (n.z < -1.0) {
		n.z = -1.0f;
	    } else if (n.z*this.coeff > 1.0) {
		n.z = 1.0f;
	    }
	    n.dx /= 1.1;
	    n.dy /= 1.1;
	    n.dz /= 1.1;
	}



   Transform3D temp = new Transform3D();
   Transform3D temp_lb = new Transform3D();
   for(int i = 0;i < this.nodes.length; i++){

     temp.set(new Vector3f(this.nodes[i].x,this.nodes[i].y,this.nodes[i].z));
     temp_lb.set(new Vector3f(this.nodes[i].x,this.nodes[i].y+0.05f,this.nodes[i].z));
     this.nodes[i].tg.setTransform(temp);
     this.nodes[i].lg.setTransform(temp_lb);


    }
  //Moving edges:
  if(!this.c_box.isSelected()){
    for (int j = 0; j < this.edges.length; j++) {
      fr = new Point3f(nodes[edges[j].from].x, nodes[edges[j].from].y,
		       nodes[edges[j].from].z);
      to = new Point3f(nodes[edges[j].to].x, nodes[edges[j].to].y,
		       nodes[edges[j].to].z);
      float elen = fr.distance(to);
      double ratio = elen / (edges[j].len * coeff);
      if (ratio < 0.5)
	this.edges[j].la.setColors(0, stress[4]);
      else if (ratio < 0.90)
	this.edges[j].la.setColors(0, stress[3]);
      else if (ratio < 1.10)
	this.edges[j].la.setColors(0, stress[0]);
      else if (ratio < 1.5)
	this.edges[j].la.setColors(0, stress[2]);
      else
	this.edges[j].la.setColors(0, stress[1]);

      Point3f[] parr = {
	  fr, to};
      this.edges[j].la.setCoordinates(0, parr);
    }
   }else{
   //Move kedges
   for (int k = 0; k < this.kedges.length; k++) {
     MoveEdge(k,1);
   }
   }
   fr = null;
   to = null;
   }


/**
* Routine for reading distances from a file
* @param  Input  file for reading
*/
    protected void ReadLines(RandomAccessFile Input) {
	String Line;
	try {
	    int rows = 0;
	    float max = 0;
	    float DMATRIX[][] = {{0},{0}};

	    while ((Line = Input.readLine()) != null) {
		if (Line.indexOf(">") >= 0) {
                    String[] cols = Line.split("\t");
                    if(Line.indexOf(">N") >= 0){
                      MATRIX = new int[cols.length - 2][cols.length - 2];
                      DMATRIX = new float[cols.length - 2][cols.length - 2];
                      NAMES = new String[cols.length - 2];
                      for (int i = 0; i < NAMES.length; i++) {
                        NAMES[i] = cols[i + 2];
                      }
                    }else if(Line.indexOf(">D")>=0){
                      DESCS = new String[cols.length - 2];
                      for (int i = 0; i < NAMES.length; i++) {
                        DESCS[i] = cols[i + 2];
                      }

                    }
                    continue;
		}
		String[] rws = Line.split("\t");
		for (int j = 3 + rows; j < rws.length; j++) {
		    DMATRIX[rows][j - 2] = Float.parseFloat(rws[j]);
		    if(DMATRIX[rows][j-2] > max)
		      max = DMATRIX[rows][j-2];

		}
		rows++;
	    }
	    float coeff = 500.0f/max;
	    for(int m = 0; m < DMATRIX.length; m++){
	     for(int n = 0; n < DMATRIX[m].length; n++)
	       MATRIX[m][n] = (int)(DMATRIX[m][n] * coeff);

	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }


/**
* Event processing routine for Opening files
*/

private void Open_actionPerformed(ActionEvent e){
     fileChooser.setDialogTitle("Open File");
	if (fileChooser.APPROVE_OPTION == fileChooser.showOpenDialog(this)) {
	    try {
		RandomAccessFile f = new RandomAccessFile(fileChooser.getSelectedFile().getPath(), "r");
		dataLoaded = false;
		ReadLines(f);
		this.c_box.setSelected(false);

		this.updateGraph(NAMES,MATRIX);
	    } catch (FileNotFoundException nfe) {
		nfe.printStackTrace();
	    }
	}
   }


/**
* Event processing routine for changing the positions of elements
*/

private void Test_actionPerformed(ActionEvent e){
  if(this.nodes == null)
     return;

   //Moving nodes:
   Transform3D temp = new Transform3D();
   for(int i = 0;i < this.nodes.length; i++){
     nodes[i].x = (float)Math.random()/2;
     nodes[i].y = (float)Math.random()/2;
     nodes[i].z = (float)Math.random()/2;

     temp.set(new Vector3f(this.nodes[i].x,this.nodes[i].y,this.nodes[i].z));
     this.nodes[i].tg.setTransform(temp);
     }


  //Moving edges:
  if(this.c_box.isSelected()){
    //MOVE knots:
    for(int n = 0;n < this.knots.length;n++){
      if(knots[n].Linked[0]>=NAMES.length){
	knots[n].x = knots[knots[n].Linked[0]-NAMES.length].x;
	knots[n].y = knots[knots[n].Linked[0]-NAMES.length].y;
	knots[n].z = knots[knots[n].Linked[0]-NAMES.length].z;
      }else{
	knots[n].x = nodes[knots[n].Linked[0]].x;
	knots[n].y = nodes[knots[n].Linked[0]].y;
	knots[n].z = nodes[knots[n].Linked[0]].z;
      }
      if(knots[n].Linked[1]>=NAMES.length){
	knots[n].x = (knots[n].x+knots[knots[n].Linked[1]-NAMES.length].x)/2;
	knots[n].y = (knots[n].y+knots[knots[n].Linked[1]-NAMES.length].y)/2;
	knots[n].z = (knots[n].z+knots[knots[n].Linked[1]-NAMES.length].z)/2;
      }else{
	knots[n].x = (knots[n].x+nodes[knots[n].Linked[1]].x)/2;
	knots[n].y = (knots[n].y+nodes[knots[n].Linked[1]].y)/2;
	knots[n].z = (knots[n].z+nodes[knots[n].Linked[1]].z)/2;
      }
     temp.set(new Vector3f(knots[n].x,knots[n].y,knots[n].z));
     this.knots[n].tg.setTransform(temp);
    }
    //Move Kedges:
    for(int k = 0;k < this.kedges.length;k++)
      MoveEdge(k,1);
  }else{
    for(int j = 0;j < this.edges.length;j++)
    MoveEdge(j,0);
  }
}
/**
 * Small function for moving edges (kedges) on the graph
*/
private void MoveEdge(int e,int etype){
    int f,t;
    try{
    Point3f [] parr = new Point3f [2];

    if (etype == 1){
     f = kedges[e].from;
     t = kedges[e].to;
    }else{
     f = edges[e].from;
     t = edges[e].to;
    }
   parr[0]=(f>=NAMES.length) ? new Point3f(knots[f-NAMES.length].x,knots[f-NAMES.length].y,knots[f-NAMES.length].z) : new Point3f(nodes[f].x,nodes[f].y,nodes[f].z);
   parr[1]=(t>=NAMES.length) ? new Point3f(knots[t-NAMES.length].x,knots[t-NAMES.length].y,knots[t-NAMES.length].z) : new Point3f(nodes[t].x,nodes[t].y,nodes[t].z);

    if(etype == 1)
       kedges[e].la.setCoordinates(0,parr);
    else
       edges[e].la.setCoordinates(0,parr);
    }catch(java.lang.NullPointerException npe){
      npe.printStackTrace();
      return;
    }
   }


/**
* Event processing routine - alternative edge visibility rule (NJ)
*/
private void Cbox_actionPerformed(ActionEvent e){
    this.dataLoaded = true;
Update_Visibility();
}

/**
* Event processing routine - alternative edge visibility rule (NJ)
*/
private void Dbox_actionPerformed(ActionEvent e){

if(this.d_box.isSelected()){
 for (int i = 0; i<this.nodes.length; i++){
   if(this.DESCS[i].length() > description_length)
     this.DESCS[i] = this.DESCS[i].substring(0,description_length);
   else{
     while(DESCS[i].length() < description_length)
       DESCS[i] += " ";
   }

   this.nodes[i].lbl.setString(this.DESCS[i]);
 }
}else{
 for (int i = 0; i<this.nodes.length; i++){

   this.nodes[i].lbl.setString(this.nodes[i].id);
 }
}
}


/**
* Event processing routine for changing the edges' visibility parameters
*/
private void VSlider_stateChanged(ChangeEvent e){
Update_Visibility();
}

/**
* Event processing routine for changing the edges' visibility parameters
*/
private void DSlider_stateChanged(ChangeEvent e){
float num = this.DSlider.getValue()/5.0f;
this.coeff = this.refcf*num;
}

/**
* A routine updating the visibility of connecting edges
*/
private void Update_Visibility(){

  if(this.edges == null||this.dataLoaded == false)
    return;
  for(int c = 0;c < this.edges.length;c++){
    if(this.c_box.isSelected())
      this.edges[c].line.getAppearance().getRenderingAttributes().setVisible(false);
    else{
     if(this.edges[c].len > this.VSlider.getValue()){
	this.edges[c].line.getAppearance().getRenderingAttributes().setVisible(false);
	}
     else
       this.edges[c].line.getAppearance().getRenderingAttributes().setVisible(true);
    }
  }

   for(int k = 0;k < this.kedges.length;k++){
    if(this.c_box.isSelected()){
      if(this.kedges[k].len > this.VSlider.getValue()){
	this.kedges[k].line.getAppearance().getRenderingAttributes().setVisible(false);
	}
    else
      this.kedges[k].line.getAppearance().getRenderingAttributes().setVisible(true);
   }else
      this.kedges[k].line.getAppearance().getRenderingAttributes().setVisible(false);
   }

   for(int sn = 0;sn < this.knots.length; sn++){
    if(this.c_box.isSelected())
      this.knots[sn].getAppearance().getRenderingAttributes().setVisible(true);
    else
      this.knots[sn].getAppearance().getRenderingAttributes().setVisible(false);
   }
}


/**
* Special subroutine for building Neighbor-Joining tree - implemented a standard
* algorithm for 'classic' NJ tree
*/
private void buildTree(){
   if(this.dataLoaded && this.kedges[0] != null || this.NAMES.length == 1){return;}
   int cnode, ckedge; cnode = NAMES.length; ckedge = 0;//current ids for knot and kedge
   boolean connected = false;//probably we don't need it (or do we?)
   int cluster[][] = new int[NAMES.length][NAMES.length];//ids of nodes' pairs already connected
   Hashtable subs    = new Hashtable();//register topnodes here
   int min,x,y; min = 0; x = 0; y = 0;

   while(!connected){
   //find next min distance between unconnected nodes
   for(int a = 0; a < NAMES.length; a++){
   for(int b = a+1; b < NAMES.length; b++){
     if (cluster[a][b] == 0 && cluster[b][a] == 0) {
       //Our assumption here that no value in MATRIX equal zero
       if ( (min == 0 || MATRIX[a][b] <= min) && MATRIX[a][b] > 0) {
	 min = MATRIX[a][b];
	 x = a;
	 y = b;
       }
     }
   }
   }
	 //terminate it right away if we have only two nodes
	 if(NAMES.length == 2){
	   kedges[0] = new Edge(x,y);
	   kedges[0].len  = MATRIX[x][y];
	   break;//we don't really need to set the flag here
	 }
       //Now we have here a new pair of nodes not previously connected
       //a: check if one/both of them in subclusters
       int from,to,f_off,t_off;
       cluster[x][y] = 1; //mark that we connected these two nodes
       min = 0;//as we set the cluster flag, min should be reset
       //kedge should be generated b/w topnodes (if nodes in subclusters already)
       //or between nodes, or between node nad subcluster
       from = (subs.get(""+x) != null) ? Integer.parseInt(subs.get(""+x).toString()) : x;
       to   = (subs.get(""+y) != null) ? Integer.parseInt(subs.get(""+y).toString()) : y;
       if(from == to)
	  continue;

       f_off = (from >=NAMES.length) ? knots[from-NAMES.length].offset : 0;
       t_off = (to   >=NAMES.length) ? knots[to  -NAMES.length].offset : 0;
       //simple check: if we have knots available, connect with subnode otherwize -
       //terminate the tree with last kedge
       try{
	if(knots[cnode-NAMES.length] == null){
	 kedges[ckedge] = new Edge(from,cnode);
	 kedges[ckedge++].len = Math.round(MATRIX[x][y]/2-f_off);
	 kedges[ckedge] = new Edge(cnode,to);
	 kedges[ckedge++].len = Math.round(MATRIX[x][y]-t_off);
	 knots[cnode-NAMES.length] = new SubNode(0.01f,from,to);
	 //Faulty block:
	 for(int l = 0; l<knots[cnode-NAMES.length].Linked.length; l++){
	   if(knots[cnode-NAMES.length].Linked[l] == x || knots[cnode-NAMES.length].Linked[l] == y)
	    subs.put(""+knots[cnode-NAMES.length].Linked[l],""+cnode);
	   else{
	     Enumeration en = subs.keys();
	     while(en.hasMoreElements()){
	       String tn = en.nextElement().toString();
	       if(Integer.parseInt(subs.get(tn).toString()) == knots[cnode-NAMES.length].Linked[l]){
		 subs.remove(tn);
		 subs.put(tn,""+cnode);
	       }
	     }
	   }
	 }
	 cnode++;
	 continue;
	}
       }catch(java.lang.ArrayIndexOutOfBoundsException e){
	kedges[ckedge] = new Edge(to,from);
	kedges[ckedge].len = MATRIX[x][y]-(f_off+t_off);
	break;
       }

   connected = true;//earlier code should provide 'continue' to avoid setting this flag
   }
  }

private void jbInit() throws Exception {
    this.addComponentListener(new tb3d_this_componentAdapter(this));
  }

//Here we check with the data from the control panel
void this_componentShown(ComponentEvent e) {
if(tb3d_controls.FRESH){
      this.dataLoaded = false;
      this.MATRIX = tb3d_controls.MATRIX;
      this.NAMES  = tb3d_controls.NAMES;
      this.DESCS  = tb3d_controls.DESCS;

      this.updateGraph(this.NAMES,this.MATRIX);
      tb3d_controls.FRESH = false;
    }
}


class tb3d_this_componentAdapter extends java.awt.event.ComponentAdapter {
  tb3d adaptee;

  tb3d_this_componentAdapter(tb3d adaptee) {
    this.adaptee = adaptee;
  }
  public void componentShown(ComponentEvent e) {
    adaptee.this_componentShown(e);
  }
}
}

